@@ -53,6 +53,14 @@ struct sock_list;
 struct ub_packed_rrset_key;
 struct regional;
 
+/** List head for strlist processing, used for append operation. */
+struct config_strlist_head {
+	/** first in list of text items */
+	struct config_strlist* first;
+	/** last in list of text items */
+	struct config_strlist* last;
+};
+
 /**
  * The configuration options.
  * Strings are malloced.
@@ -91,6 +99,12 @@ struct config_file {
 	int tcp_mss;
 	/** maximum segment size of tcp socket for outgoing queries */
 	int outgoing_tcp_mss;
+	/** tcp idle timeout, in msec */
+	int tcp_idle_timeout;
+	/** do edns tcp keepalive */
+	int do_tcp_keepalive;
+	/** tcp keepalive timeout, in msec */
+	int tcp_keepalive_timeout;
 
 	/** private key file for dnstcp-ssl service (enabled if not NULL) */
 	char* ssl_service_key;
@@ -102,8 +116,10 @@ struct config_file {
 	int ssl_upstream;
 	/** cert bundle for outgoing connections */
 	char* tls_cert_bundle;
+	/** should the system certificate store get added to the cert bundle */
+	int tls_win_cert;
 	/** additional tls ports */
-	struct config_strlist* additional_tls_port;
+	struct config_strlist* tls_additional_port;
 
 	/** outgoing port range number of ports (per thread) */
 	int outgoing_num_ports;
@@ -144,7 +160,7 @@ struct config_file {
 	/** the target fetch policy for the iterator */
 	char* target_fetch_policy;
 	/** percent*10, how many times in 1000 to pick low rtt destinations */
-	int low_rtt_pct;
+	int low_rtt_permil;
 	/** what time in msec is a low rtt destination */
 	int low_rtt;
 
@@ -204,6 +220,9 @@ struct config_file {
 	/** use default localhost donotqueryaddr entries */
 	int donotquery_localhost;
 
+	/** list of tcp connection limitss, linked list */
+	struct config_str2list* tcp_connection_limits;
+
 	/** harden against very small edns buffer sizes */
 	int harden_short_bufsize;
 	/** harden against very large query sizes */
@@ -258,6 +277,10 @@ struct config_file {
 	int log_queries;
 	/** log replies with one line per reply */
 	int log_replies;
+	/** log every local-zone hit **/
+	int log_local_actions;
+	/** log servfails with a reason */
+	int log_servfail;
 	/** log identity to report */
 	char* log_identity;
 
@@ -316,6 +339,10 @@ struct config_file {
 	int ignore_cd;
 	/** serve expired entries and prefetch them */
 	int serve_expired;
+	/** serve expired entries until TTL after expiration */
+	int serve_expired_ttl;
+	/** reset serve expired TTL after failed update attempt */
+	int serve_expired_ttl_reset;
 	/** nsec3 maximum iterations per key size, string */
 	char* val_nsec3_key_iterations;
 	/** autotrust add holddown time, in seconds */
@@ -372,11 +399,11 @@ struct config_file {
 	/** remote control section. enable toggle. */
 	int remote_control_enable;
 	/** the interfaces the remote control should listen on */
-	struct config_strlist* control_ifs;
+	struct config_strlist_head control_ifs;
+	/** if the use-cert option is set */
+	int control_use_cert;
 	/** port number for the control port */
 	int control_port;
-	/** use certificates for remote control */
-	int remote_control_use_cert;
 	/** private key file for server */
 	char* server_key_file;
 	/** certificate file for server */
@@ -409,6 +436,8 @@ struct config_file {
 
 	/* Synthetize all AAAA record despite the presence of an authoritative one */
 	int dns64_synthall;
+	/** ignore AAAAs for these domain names and use A record anyway */
+	struct config_strlist* dns64_ignore_aaaa;
 
 	/** true to enable dnstap support */
 	int dnstap;
@@ -551,6 +580,8 @@ struct config_stub {
 	int isfirst;
 	/** use SSL for queries to this stub */
 	int ssl_upstream;
+	/*** no cache */
+	int no_cache;
 };
 
 /**
@@ -651,14 +682,6 @@ struct config_strbytelist {
 	size_t str2len;
 };
 
-/** List head for strlist processing, used for append operation. */
-struct config_strlist_head {
-	/** first in list of text items */
-	struct config_strlist* first;
-	/** last in list of text items */
-	struct config_strlist* last;
-};
-
 /**
  * Create config file structure. Filled with default values.
  * @return: the new structure or NULL on memory error.
@@ -892,6 +915,10 @@ void config_delview(struct config_view* p);
  */
 void config_delviews(struct config_view* list);
 
+/** check if config for remote control turns on IP-address interface
+ * with certificates or a named pipe without certificates. */
+int options_remote_is_address(struct config_file* cfg);
+
 /**
  * Convert 14digit to time value
  * @param str: string of 14 digits
@@ -1059,12 +1086,20 @@ void errinf_dname(struct module_qstate* qstate, const char* str,
 	uint8_t* dname);
 
 /**
- * Create error info in string
+ * Create error info in string.  For validation failures.
  * @param qstate: query state.
  * @return string or NULL on malloc failure (already logged).
  *    This string is malloced and has to be freed by caller.
  */
-char* errinf_to_str(struct module_qstate* qstate);
+char* errinf_to_str_bogus(struct module_qstate* qstate);
+
+/**
+ * Create error info in string.  For other servfails.
+ * @param qstate: query state.
+ * @return string or NULL on malloc failure (already logged).
+ *    This string is malloced and has to be freed by caller.
+ */
+char* errinf_to_str_servfail(struct module_qstate* qstate);
 
 /**
  * Used during options parsing
